/*! 
  @file main.cpp
	
  @brief Bulletɂ镨V~[V
		 btCompoundShapeɂ镡`̃eXg
 
  @author Makoto Fujisawa
  @date 2013-03
*/

#pragma comment (lib, "glew32.lib")

#ifdef _DEBUG
#pragma comment (lib, "LinearMath_vs2010_debug.lib")
#pragma comment (lib, "BulletCollision_vs2010_debug.lib")
#pragma comment (lib, "BulletDynamics_vs2010_debug.lib")
#else
#pragma comment (lib, "LinearMath_vs2010.lib")
#pragma comment (lib, "BulletCollision_vs2010.lib")
#pragma comment (lib, "BulletDynamics_vs2010.lib")
#endif

//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "utils.h"

//-----------------------------------------------------------------------------
// MARK:`/萔
//-----------------------------------------------------------------------------
const GLfloat RX_LIGHT0_POS[4] = {  2.0f, 4.0f, 1.0f, 0.0f };
const GLfloat RX_LIGHT1_POS[4] = { -1.0f, -10.0f, -1.0f, 0.0f };

const GLfloat RX_LIGHT_DIFF[4] = { 1.0f, 1.0f, 1.0f, 1.0f };
const GLfloat RX_LIGHT_SPEC[4] = { 0.7f, 0.6f, 0.6f, 1.0f };
const GLfloat RX_LIGHT_AMBI[4] = { 0.3f, 0.3f, 0.3f, 1.0f };

const GLfloat RX_FOV = 45.0f;

const btScalar CUBE_SCALE = 0.4;
const btScalar GROUND_HEIGHT = -2.0;

const double DT = 0.01;					//!< ԃXebvt

// V[
enum{
	RXS_CAPSULES = 0, 
	RXS_CONES, 
	RXS_BOXES, 
};


//-----------------------------------------------------------------------------
// MARK:O[oϐ
//-----------------------------------------------------------------------------
//! EBhETCY
int g_iWinW = 720;
int g_iWinH = 720;
int g_iMouseButton = -1;	//!< }EX{^̏
bool g_bIdle;				//!< Aj[VtO

//! gbN{[
rxTrackball g_tbView;

// VhE}bsO
rxShadowMap g_ShadowMap;
int g_iShadowMapSize = 512;

// Bullet
btDynamicsWorld* g_pDynamicsWorld;	//!< Bullet[h

// Xebv
int g_iStep = 0;

// }EXsbN
btVector3 g_vPickPos;
btRigidBody *g_pPickBody = 0;
btPoint2PointConstraint *g_pPickConstraint = 0;
double g_fPickDist = 0.0;

int g_iScene = RXS_CAPSULES;


//-----------------------------------------------------------------------------
// MARK:Bulletp֐
//-----------------------------------------------------------------------------
/*!
 * Bullet(btRigidBody)̍쐬
 * @param[in] mass 
 * @param[in] init_tras ʒuEp
 * @param[in] shape `
 * @param[in] world Bullet[h(0Ȃ΃[hւ̓o^͍sȂ)
 * @return 쐬btRigidBody
 */
btRigidBody* CreateRigidBody(double mass, const btTransform& init_trans, btCollisionShape* shape, btDynamicsWorld* world = 0)
{
	btAssert((!shape || shape->getShapeType() != INVALID_SHAPE_PROXYTYPE));

	// ʂ0ȂΐÓI(static)IuWFNgƂĐݒC
	bool isDynamic = (mass != 0.0);

	btVector3 inertia(0, 0, 0);
	if(isDynamic)
		shape->calculateLocalInertia(mass, inertia);

	btDefaultMotionState* motion_state = new btDefaultMotionState(init_trans);

	btRigidBody::btRigidBodyConstructionInfo rb_info(mass, motion_state, shape, inertia);

	btRigidBody* body = new btRigidBody(rb_info);

	if(world){
		world->addRigidBody(body);
	}

	return body;
}

/*!
 * BulletԂɃ{bNX`zu
 * @param[in] pos ʒu
 * @param[in] len Ӓ
 * @param[in] mass 
 * @param[in] world Bullet[h(0Ȃ΃[hւ̓o^͍sȂ)
 * @return 쐬btRigidBody
 */
btRigidBody* SetBox(btVector3 &pos, btVector3 &len, double mass, btDynamicsWorld* world = 0)
{
	btVector3 half_ext = 0.5*len;

	// ̃IuWFNg
	btBoxShape *shape = new btBoxShape(half_ext);

	// ʒuEp
	btQuaternion qrot(0, 0, 0, 1);
	btTransform trans;	// ̃IuWFNg̈ʒupi[ϐ(s)
	trans.setIdentity();// ʒups̏
	trans.setOrigin(pos);
	trans.setRotation(qrot);	// lsɕϊĎpsɊ|킹

	// ̃IuWFNg
	btRigidBody* body = CreateRigidBody(mass, trans, shape, world);

	// 蔲h~pSwept sphere̐ݒ(CCD:Continuous Collision Detection)
	btScalar min_extent = RX_MIN3(half_ext[0], half_ext[1], half_ext[2]);
	body->setCcdMotionThreshold(min_extent);
	body->setCcdSweptSphereRadius(0.05*min_extent);

	if(mass <= 1e-10){
		// KinematicIuWFNgƂĐݒ(stepSimulationĂ^̌vZsȂ悤ɂ)
		body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT);
		// ɃX[vԂɂ
		body->setActivationState(DISABLE_DEACTIVATION);
	}

	return body;
}

/*!
 * BulletԂɗ̂zu
 * @param[in] pos ʒu
 * @param[in] len Ӓ
 * @param[in] mass 
 * @param[in] world Bullet[h(0Ȃ΃[hւ̓o^͍sȂ)
 * @return 쐬btRigidBody
 */
btRigidBody* SetCube(btVector3 &pos, double len, double mass, btDynamicsWorld* world = 0)
{
	return SetBox(pos, btVector3(len, len, len), mass, world);
}

/*!
 * BulletԂɋ̂zu
 * @param[in] pos ʒu
 * @param[in] rad a
 * @param[in] mass 
 * @param[in] world Bullet[h(0Ȃ΃[hւ̓o^͍sȂ)
 * @return 쐬btRigidBody
 */
btRigidBody* SetSphere(btVector3 &pos, double rad, double mass, btDynamicsWorld* world = 0)
{
	// ̃IuWFNg
	btSphereShape *shape = new btSphereShape((btScalar)rad);

	// ʒuEp
	btQuaternion qrot(0, 0, 0, 1);
	qrot.setEuler(0, 0, 0);
	btTransform trans;	// ̃IuWFNg̈ʒupi[ϐ(s)
	trans.setIdentity();// ʒups̏
	trans.setOrigin(pos);
	trans.setRotation(qrot);	// lsɕϊĎpsɊ|킹

	// ̃IuWFNg
	btRigidBody* body = CreateRigidBody(mass, trans, shape, world);

	// 蔲h~pSwept sphere̐ݒ(CCD:Continuous Collision Detection)
	body->setCcdMotionThreshold(rad);
	body->setCcdSweptSphereRadius(0.05*rad);

	if(mass <= 1e-10){
		// KinematicIuWFNgƂĐݒ(stepSimulationĂ^̌vZsȂ悤ɂ)
		body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT);
		// ɃX[vԂɂ
		body->setActivationState(DISABLE_DEACTIVATION);
	}

	return body;
}


/*!
 * BulletԂɉ~zu(x)
 * @param[in] pos ʒu
 * @param[in] rad a
 * @param[in] len 
 * @param[in] mass 
 * @param[in] world Bullet[h(0Ȃ΃[hւ̓o^͍sȂ)
 * @return 쐬btRigidBody
 */
btRigidBody* SetCylinder(btVector3 &pos, double rad, double len, double mass, btDynamicsWorld* world = 0)
{
	// ̃IuWFNg
	btVector3 half_extents(0.5*len, rad, rad);
	btCylinderShape *shape = new btCylinderShapeX(half_extents);
	shape->setMargin(0.005);

	// ʒuEp
	btQuaternion qrot(0, 0, 0, 1);
	qrot.setEuler(0, 0, 0);
	btTransform trans;	// ̃IuWFNg̈ʒupi[ϐ(s)
	trans.setIdentity();// ʒups̏
	trans.setOrigin(pos);
	trans.setRotation(qrot);	// lsɕϊĎpsɊ|킹

	// ̃IuWFNg
	btRigidBody* body = CreateRigidBody(mass, trans, shape, world);

	// 蔲h~pSwept sphere̐ݒ(CCD:Continuous Collision Detection)
	body->setCcdMotionThreshold(rad);
	body->setCcdSweptSphereRadius(0.05*rad);

	if(mass <= 1e-10){
		// KinematicIuWFNgƂĐݒ(stepSimulationĂ^̌vZsȂ悤ɂ)
		body->setCollisionFlags(body->getCollisionFlags() | btCollisionObject::CF_KINEMATIC_OBJECT);
		// ɃX[vԂɂ
		body->setActivationState(DISABLE_DEACTIVATION);
	}

	return body;
}

/*!
 * ̃IuWFNg̒ǉ
 */
void SetRigidBodies(void)
{
	btTransform trans;	// ̃IuWFNg̈ʒupi[ϐ(s)
	trans.setIdentity();// ʒups̏

	// n(0x-zʏŕׂ̂ŕ\)
	btBoxShape *ground_shape = new btBoxShape(btVector3(20, 0.5*CUBE_SCALE, 20));	// `
	trans.setOrigin(btVector3(0, GROUND_HEIGHT-0.5*CUBE_SCALE, 0));	// ̖ʂGROUND_HEIGHTɂȂ悤ɐݒ
	btRigidBody* ground = CreateRigidBody(0.0, trans, ground_shape, 0);	// Body
	ground->setFriction(0.8);
	g_pDynamicsWorld->addRigidBody(ground);

	if(g_iScene == RXS_CAPSULES){
		// JvZgݍ킹`
		btCompoundShape *compound = new btCompoundShape;

		// JvZ`2`
		btScalar r = 0.1, l = 0.5;
		btCapsuleShape *capsule1 = new btCapsuleShape(r, l);
		btCapsuleShape *capsule2 = new btCapsuleShape(r, 2.0*l);

		trans.setIdentity();

		// JvZ`1btCompoundShapeɎqm[hƂĒǉ
		compound->addChildShape(trans, capsule1);

		// JvZ`2btCompoundShape90x]Ďqm[hƂĒǉ
		trans.setOrigin(btVector3(0, 0.5*l, 0));
		trans.setRotation(btQuaternion(0, 0, 0.5*RX_PI));
		compound->addChildShape(trans, capsule2);

		// btCompoundShapeAABBČvZ
		compound->recalculateLocalAabb();

		// btRigidBody̍쐬ƃ[hւ̒ǉ
		trans.setIdentity();
		trans.setOrigin(btVector3(0, GROUND_HEIGHT+1.0, 0));
		btRigidBody* body1 = CreateRigidBody(0.2, trans, compound, 0);	// Body
		g_pDynamicsWorld->addRigidBody(body1);
	}
	else if(g_iScene == RXS_CONES){
		// ~gݍ킹`
		btCompoundShape *compound = new btCompoundShape;

		// ~`3쐬
		btScalar r = 0.2, l = 0.5;
		btConeShape *cone1 = new btConeShape(r, l);	// ftHgY (btConeShapeX  btConeShapeZ )
		btConeShape *cone2 = new btConeShape(1.5*r, l);
		btConeShape *cone3 = new btConeShape(2.0*r, l);

		// N炵ȂbtCompoundShapeɎqm[hƂĒǉ
		trans.setIdentity();
		compound->addChildShape(trans, cone1);
		trans.setOrigin(btVector3(0, -0.5*l, 0));
		compound->addChildShape(trans, cone2);
		trans.setOrigin(btVector3(0, -1.0*l, 0));
		compound->addChildShape(trans, cone3);

		// btCompoundShapeAABBČvZ
		compound->recalculateLocalAabb();

		// btRigidBody̍쐬ƃ[hւ̒ǉ
		trans.setIdentity();
		trans.setOrigin(btVector3(0, GROUND_HEIGHT+1.0, 0));
		btRigidBody* body1 = CreateRigidBody(0.2, trans, compound, 0);	// Body
		g_pDynamicsWorld->addRigidBody(body1);
	}
	else if(g_iScene == RXS_BOXES){
		// ̂gݍ킹`
		btCompoundShape *compound = new btCompoundShape;

		// ̌`쐬
		btScalar hl = 0.1;
		btBoxShape *box = new btBoxShape(btVector3(hl, hl, hl));

		// sړȂbtCompoundShapeɎqm[hƂĒǉ
		trans.setIdentity();
		compound->addChildShape(trans, box);
		trans.setOrigin(btVector3(0, -2*hl, 0));
		compound->addChildShape(trans, box);
		trans.setOrigin(btVector3(0, -4*hl, 0));
		compound->addChildShape(trans, box);
		trans.setOrigin(btVector3(2*hl, -4*hl, 0));
		compound->addChildShape(trans, box);

		// btCompoundShapeAABBČvZ
		compound->recalculateLocalAabb();

		// btRigidBody̍쐬ƃ[hւ̒ǉ
		trans.setIdentity();
		trans.setOrigin(btVector3(0, GROUND_HEIGHT+1.0, 0));
		btRigidBody* body1 = CreateRigidBody(0.2, trans, compound, 0);	// Body
		g_pDynamicsWorld->addRigidBody(body1);
	}

}

/*!
 * Bullet
 */
void InitBullet(void)
{
	const btScalar CUBE_HALF_EXTENTS = 0.5;
	const btScalar EXTRA_HEIGHT = 0.0;

	// Փˌo@̑I(ftHgI)
	btDefaultCollisionConfiguration *config = new btDefaultCollisionConfiguration();
	btCollisionDispatcher *dispatcher = new btCollisionDispatcher(config);

	// u[htF[Y@̐ݒ(Dynamic AABB tree method)
	//btDbvtBroadphase *broadphase = new btDbvtBroadphase();

	// u[htF[Y@̐ݒ(Sweep and Prune method)
	btAxisSweep3 *broadphase = new btAxisSweep3(btVector3(-100, -100, -100), btVector3(100, 100, 100));

	// S(̊ԃN)̃\oݒ
	btSequentialImpulseConstraintSolver* solver = new btSequentialImpulseConstraintSolver();

	// Bullet̃[h쐬
	g_pDynamicsWorld = new btDiscreteDynamicsWorld(dispatcher, broadphase, solver, config);

	// d͉x̐ݒ(OpenGLɍ킹y㉺ɂ)
	g_pDynamicsWorld->setGravity(btVector3(0, -9.8, 0));

	SetRigidBodies();
}


/*!
 * ݒ肵Bullet̍̃IuWFNgC[h̔j
 */
void CleanBullet(void)
{
	// S̔j
	for(int i = g_pDynamicsWorld->getNumConstraints()-1; i>=0 ;i--){
		btTypedConstraint* constraint = g_pDynamicsWorld->getConstraint(i);
		g_pDynamicsWorld->removeConstraint(constraint);
		delete constraint;
	}

	// ̃IuWFNg̔j
	for(int i = g_pDynamicsWorld->getNumCollisionObjects()-1; i >= 0; --i){
		btCollisionObject* obj = g_pDynamicsWorld->getCollisionObjectArray()[i];
		btRigidBody* body = btRigidBody::upcast(obj);
		if(body && body->getMotionState()){
			delete body->getMotionState();
		}

		btCollisionShape* shape = obj->getCollisionShape();
		delete shape;

		g_pDynamicsWorld->removeCollisionObject( obj );
		delete obj;
	}

	// [hj
	delete g_pDynamicsWorld->getBroadphase();
	delete g_pDynamicsWorld;
}



//-----------------------------------------------------------------------------
// `֐
//-----------------------------------------------------------------------------
/*!
 * eϊ
 */
void Projection(void)
{
	gluPerspective(RX_FOV, (double)g_iWinW/(double)g_iWinH, 0.01f, 1000.0f);
}

/*!
 * Bullet̏Փˌ``
 * @param[in] shape Փˌ`
 */
void DrawBulletShape(const btCollisionShape *shape)
{
	int shapetype = shape->getShapeType();

	glPushMatrix();

	// `̎ނƂɕ`
	if(shapetype == BOX_SHAPE_PROXYTYPE){
		// {bNX`
		const btBoxShape* box = static_cast<const btBoxShape*>(shape);
		btVector3 half_extent = box->getHalfExtentsWithMargin();
		glScaled(2*half_extent[0], 2*half_extent[1], 2*half_extent[2]);
		glutSolidCube(1.0);
	}
	else if(shapetype == SPHERE_SHAPE_PROXYTYPE){
		// `
		const btSphereShape* sphere = static_cast<const btSphereShape*>(shape);
		double rad = sphere->getRadius();
		glutSolidSphere(rad, 32, 32);
	}
	else if(shapetype == CYLINDER_SHAPE_PROXYTYPE){
		// ~`
		const btCylinderShape* cylinder = static_cast<const btCylinderShape*>(shape);
		double rad = cylinder->getRadius();
		int up_axis = cylinder->getUpAxis();
		double len = cylinder->getHalfExtentsWithMargin()[up_axis]*2;
		DrawCylinder(rad, len, up_axis, 8);
	}
	else if(shapetype == CAPSULE_SHAPE_PROXYTYPE){
		// JvZ`
		const btCapsuleShape* capsule = static_cast<const btCapsuleShape*>(shape);
		double rad  = capsule->getRadius();
		int up_axis = capsule->getUpAxis();
		double len  = capsule->getHalfHeight()*2;
		DrawCapsule(rad, len, up_axis, 16);
	}
	else if(shapetype == CONE_SHAPE_PROXYTYPE){
		// ~`
		const btConeShape* cone = static_cast<const btConeShape*>(shape);
		float rad   = cone->getRadius()+cone->getMargin();
		float len   = cone->getHeight()+cone->getMargin();
		int up_axis = cone->getConeUpIndex();
		DrawCone(rad, len, up_axis, 16);
	}
	else if(shapetype == COMPOUND_SHAPE_PROXYTYPE){
		// `
		const btCompoundShape* compound = static_cast<const btCompoundShape*>(shape);
		int num_child = compound->getNumChildShapes();
		btScalar mc[16];
		for(int j = 0; j < num_child; ++j){
			// Compound Shape\child shapeċAIDrawBulletShapeɓn
			compound->getChildTransform(j).getOpenGLMatrix(mc);
			glPushMatrix();
			glMultMatrixf(mc);
			DrawBulletShape(compound->getChildShape(j));
			glPopMatrix();
		}
	}
	else{
		glutSolidSphere(1.0, 10, 10);
	}

	glPopMatrix();
}


/*!
 * Bullet̃IuWFNg̕`V[`
 */
void DrawBulletObjects(void)
{
	static const GLfloat difr[] = { 1.0, 0.4, 0.4, 1.0 };	// gUF : 
	static const GLfloat difg[] = { 0.4, 0.6, 0.4, 1.0 };	// gUF : 
	static const GLfloat difb[] = { 0.4, 0.4, 1.0, 1.0 };	// gUF : 
	static const GLfloat spec[] = { 0.3, 0.3, 0.3, 1.0 };	// ʔːF
	static const GLfloat ambi[] = { 0.1, 0.1, 0.1, 1.0 };	// 

	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	glDisable(GL_COLOR_MATERIAL);

	glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
	glMaterialfv(GL_FRONT, GL_AMBIENT,  ambi);
	glMaterialf(GL_FRONT, GL_SHININESS, 50.0f);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glColor3f(0.0, 0.0, 1.0);

	glDisable(GL_CULL_FACE);

	if(g_pDynamicsWorld){
		btScalar m[16];
		btMatrix3x3	rot;
		rot.setIdentity();

		// Bullet[h獄̃IuWFNg擾OpenGLŕ`
		const int n = g_pDynamicsWorld->getNumCollisionObjects();	// IuWFNg̎擾
		for(int i = 0; i < n; ++i){

			// btCollisionObject  btRigidBodyւ̃LXgō̃IuWFNg擾
			btCollisionObject* obj = g_pDynamicsWorld->getCollisionObjectArray()[i];
			btRigidBody* body = btRigidBody::upcast(obj);

			if(body && body->getMotionState()){
				// btRigidBodyMotion State擾āCOpenGL̕ϊsƂĈʒuEp𓾂
				btDefaultMotionState* ms = (btDefaultMotionState*)body->getMotionState();
				ms->m_graphicsWorldTrans.getOpenGLMatrix(m);
				rot = ms->m_graphicsWorldTrans.getBasis();
			}
			else{
				obj->getWorldTransform().getOpenGLMatrix(m);
				rot = obj->getWorldTransform().getBasis();
			}
			
			if(body == g_pPickBody){
				// }EXőIꂽIuWFNg͐Ԃŕ`
				glMaterialfv(GL_FRONT, GL_DIFFUSE, difr);
			}
			else if(body && body->getInvMass() > RX_FEQ_EPS){
				// Dynamic Body̏ꍇ͐ŕ`
				glMaterialfv(GL_FRONT, GL_DIFFUSE, difb);
			}
			else{
				// Kinematic Body̏ꍇ͗΂ŕ`
				glMaterialfv(GL_FRONT, GL_DIFFUSE, difg);
			}

			glPushMatrix(); 
			glMultMatrixf(m);

			// `擾
			btCollisionShape* shape = obj->getCollisionShape();
			
			// ``
			DrawBulletShape(shape);
			
			glPopMatrix();

		}
	}
}


/*!
 * V[`
 */
void RenderScene(void* x = 0)
{
	// ݒ
	glLightfv(GL_LIGHT0, GL_POSITION, RX_LIGHT0_POS);

	DrawBulletObjects();
}


/*!
 * `֐
 */
void Display(void)
{
	// t[obt@ƃfvXobt@NA
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	g_tbView.Apply();	// }EXɂ]Esړ̓Kp

	// VhE}bvgĉetŃIuWFNg`
	Vec3 light_pos = Vec3(RX_LIGHT0_POS[0], RX_LIGHT0_POS[1], RX_LIGHT0_POS[2]);
	rxFrustum light = CalFrustum(90, 0.02, 50.0, g_iShadowMapSize, g_iShadowMapSize, light_pos, Vec3(0.0));
	g_ShadowMap.RenderSceneWithShadow(light, RenderScene, 0);
	//RenderScene();

	glPopMatrix();

	glutSwapBuffers();
}

/*!
 * TCYCxg֐
 * @param[in] w,h LoXTCY
 */
void Resize(int w, int h)
{
	glViewport(0, 0, w, h);
	g_tbView.SetRegion(w, h);

	g_iWinW = w;
	g_iWinH = h;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	Projection();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/*!
 * }EXCxg֐
 * @param[in] button }EX{^(GLUT_LEFT_BUTTON,GLUT_MIDDLE_BUTTON,GLUT_RIGHT_BUTTON)
 * @param[in] state }EX{^̏(GLUT_UP, GLUT_DOWN)
 * @param[in] x,y }EXW(XN[Wn)
 */
void Mouse(int button, int state, int x, int y)
{
	if(x < 0 || y < 0) return;
	g_iMouseButton = button;
	int mod = glutGetModifiers();	// SHIFT,CTRL,ALT̏Ԏ擾

	if(button == GLUT_LEFT_BUTTON){
		if(mod == GLUT_ACTIVE_SHIFT){
			if(state == GLUT_DOWN){
				double ray_from0[3], ray_to0[3];
				double init_pos[3] = {0, 0, 0};
				g_tbView.CalLocalPos(ray_from0, init_pos);

				g_tbView.GetRayTo(x, y, RX_FOV, ray_to0);

				btVector3 ray_from = btVector3(ray_from0[0], ray_from0[1], ray_from0[2]);
				btVector3 ray_to = btVector3(ray_to0[0], ray_to0[1], ray_to0[2]);

				btCollisionWorld::ClosestRayResultCallback ray_callback(ray_from, ray_to);
				g_pDynamicsWorld->rayTest(ray_from, ray_to, ray_callback);

				if(ray_callback.hasHit()){
					const btCollisionObject* obj = ray_callback.m_collisionObject;

					// ƏՓ˂
					btRigidBody* body = const_cast<btRigidBody*>(btRigidBody::upcast(obj));

					// Փ˓_W(WCgɂȂʒuW)
					btVector3 picked_pos = ray_callback.m_hitPointWorld;

					if(body && !(body->isStaticObject() || body->isKinematicObject())){

						g_pPickBody = body;
						g_vPickPos = picked_pos;

						// Iꂽ̂̍Wnł̃sbNʒu
						btVector3 local_pos = body->getCenterOfMassTransform().inverse()*picked_pos;
						g_pPickBody->setActivationState(DISABLE_DEACTIVATION);

						if(g_pPickConstraint){
							g_pDynamicsWorld->removeConstraint(g_pPickConstraint);
							delete g_pPickConstraint;
						}
						g_pPickConstraint = new btPoint2PointConstraint(*body, local_pos);
						g_pDynamicsWorld->addConstraint(g_pPickConstraint, true);

						g_pPickConstraint->m_setting.m_impulseClamp = 30.0;
						g_pPickConstraint->m_setting.m_tau = 0.001f;

						g_fPickDist = (g_vPickPos-ray_from).length();
					}

				}

			}
		}
		else{
			if(state == GLUT_DOWN){
				if(mod == GLUT_ACTIVE_ALT){
					mod = GLUT_ACTIVE_SHIFT;
				}
				g_tbView.Start(x, y, mod+1);
			}
			else if(state == GLUT_UP){
				g_tbView.Stop(x, y);
			}
		}

		if(state == GLUT_UP){
			if(g_pPickConstraint){
				g_pDynamicsWorld->removeConstraint(g_pPickConstraint);
				delete g_pPickConstraint;
				g_pPickConstraint = 0;
				g_pPickBody->forceActivationState(ACTIVE_TAG);
				g_pPickBody->setDeactivationTime(0.0);
				g_pPickBody = 0;

			}
		}
	}
	glutPostRedisplay();

}

/*!
 * [VCxg֐(}EX{^܂܃hbO)
 * @param[in] x,y }EXW(XN[Wn)
 */
void Motion(int x, int y)
{
	if(g_iMouseButton == GLUT_LEFT_BUTTON){
		if(g_pPickConstraint){
			double ray_from0[3], ray_to0[3];
			double init_pos[3] = {0, 0, 0};
			g_tbView.CalLocalPos(ray_from0, init_pos);
			g_tbView.GetRayTo(x, y, RX_FOV, ray_to0);

			btVector3 ray_from = btVector3(ray_from0[0], ray_from0[1], ray_from0[2]);
			btVector3 new_ray_to = btVector3(ray_to0[0], ray_to0[1], ray_to0[2]);

			btVector3 dir = new_ray_to-ray_from;
			dir.normalize();

			btVector3 new_pivot = ray_from+dir*g_fPickDist;
			g_pPickConstraint->setPivotB(new_pivot);

			g_vPickPos = new_pivot;
		}
		else{
			g_tbView.Motion(x, y);
		}
	}
	glutPostRedisplay();
}

/*!
 * AChCxg֐(CPUɂȂƂɎs)
 */
void Idle(void)
{
	glutPostRedisplay();
}

/*!
 * ^C}[Cxg֐(鎞ԊԊuŎs)
 */
void Timer(int value)
{
	static btVector3 dtrans = btVector3(0.0, 0.0, 0.0);
	static btQuaternion drot = btQuaternion(0.0, 0.0, RX_PI*0.01);

	if(g_bIdle && g_pDynamicsWorld){
		// BulleťvZXebvi߂
		g_iStep = g_pDynamicsWorld->stepSimulation(DT, 1);
	}
	glutPostRedisplay();
	glutTimerFunc(DT*1000 , Timer , 0);
}


/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void Keyboard(unsigned char key, int x, int y)
{
	btRigidBody* body;
	switch(key){
	case '\033':  // '\033'  ESC  ASCII R[h
		CleanBullet();
		exit(0);

	case 's':	// Aj[V؂ւ
		g_bIdle = !g_bIdle;
		break;

	case ' ':	// 1Xebvi߂
		Idle();
		break;

	case 'r':
		CleanBullet();
		InitBullet();
		break;

	case '1':
		SetSphere(btVector3(0, GROUND_HEIGHT+5.0*CUBE_SCALE, 0), 0.5*CUBE_SCALE*RXFunc::Rand(1.5, 0.5), 1.0, g_pDynamicsWorld);
		break;

	case '2':
		SetCube(btVector3(0, GROUND_HEIGHT+5.0*CUBE_SCALE, 0), CUBE_SCALE*RXFunc::Rand(1.5, 0.5), 1.0, g_pDynamicsWorld);
		break;

	case 'x':
		{
			// _ʒu王փ{[΂
			double eye_pos[3], eye_dir[3];
			double init_pos[3] = {0, 0, 0};
			double init_dir[3] = {0, 0, -1};
			g_tbView.CalLocalPos(eye_pos, init_pos);
			g_tbView.CalLocalRot(eye_dir, init_dir);
			btRigidBody *body = SetSphere(btVector3(eye_pos[0], eye_pos[1], eye_pos[2]), 0.5*CUBE_SCALE, 1.0, g_pDynamicsWorld);
			body->setLinearVelocity(30.0*btVector3(eye_dir[0], eye_dir[1], eye_dir[2]));
		}
		break;

	default:
		break;
	}
	glutPostRedisplay();
}

/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void SpecialKey(int key, int x, int y)
{
	switch(key){
	case GLUT_KEY_UP:
		break;
	case GLUT_KEY_DOWN:
		break;
	case GLUT_KEY_RIGHT:
		break;
	case GLUT_KEY_LEFT:
		break;

	case GLUT_KEY_F1:
	case GLUT_KEY_F2:
	case GLUT_KEY_F3:
		CleanBullet();
		g_iScene = key-GLUT_KEY_F1;
		InitBullet();
		break;
	}

	glutPostRedisplay();
}

/*!
 * OpenGL̏
 */
void InitGL(void)
{
	// OpenGL̃o[W`FbN
	printf("OpenGL Ver. %s\n", glGetString(GL_VERSION));
	
	// wiF
	glClearColor(0.8, 0.8, 0.9, 1.0);

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	// ̏ݒ
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  RX_LIGHT_DIFF);
	glLightfv(GL_LIGHT0, GL_SPECULAR, RX_LIGHT_SPEC);
	glLightfv(GL_LIGHT0, GL_AMBIENT,  RX_LIGHT_AMBI);
	glLightfv(GL_LIGHT0, GL_POSITION, RX_LIGHT0_POS);

	glShadeModel(GL_SMOOTH);

	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	// _̏
	g_tbView.SetScaling(-7.0);
	//g_tbView.SetTranslation(0.0, -2.0);

	// VhE}bv
	g_ShadowMap.InitShadow(g_iShadowMapSize, g_iShadowMapSize);

	// Bullet
	InitBullet();
}


/*!
 * C[`
 * @param[in] argc R}hC̐
 * @param[in] argv R}hC
 */
int main(int argc, char *argv[])
{
	// MARK:main
	glutInitWindowPosition(100, 100);
	glutInitWindowSize(g_iWinW, g_iWinH);
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow(argv[0]);

	glutDisplayFunc(Display);
	glutReshapeFunc(Resize);
	glutMouseFunc(Mouse);
	glutMotionFunc(Motion);
	glutKeyboardFunc(Keyboard);
	glutSpecialFunc(SpecialKey);
	//glutIdleFunc(Idle);
	glutTimerFunc(DT*1000, Timer, 0);
	g_bIdle = true;

	InitGL();

	glutMainLoop();

	CleanBullet();

	return 0;
}

