/*! @file utils.h
	
	@brief lXȊ֐
 
	@author Makoto Fujisawa
	@date 2012
*/

#ifndef _RX_UTILS_H_
#define _RX_UTILS_H_

#pragma warning (disable: 4101)

//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <fstream>
#include <iostream>

#include <string>
#include <GL/glew.h>
#include <GL/glut.h>

#include "rx_utility.h"		// Vec3lXȊ֐
#include "rx_trackball.h"	// ]Csړ

#include "rx_shadow_gl.h"


#include <btBulletDynamicsCommon.h>  


using namespace std;

//-----------------------------------------------------------------------------
// 萔E`
//-----------------------------------------------------------------------------
#define RX_MAX_CONTACTS 1024

inline std::ostream &operator<<(std::ostream &out, const btVector3 &a)
{
	return out << "(" << a[0] << ", " << a[1] << ", " << a[2] << ")" ;
}

inline std::istream &operator>>(std::istream &in, btVector3& a)
{
	return in >> a[0] >> a[1] >> a[2] ;
}

inline std::ostream &operator<<(std::ostream &out, const btQuaternion &a)
{
	return out << "(" << a[0] << ", " << a[1] << ", " << a[2] << ", " << a[3] << ")" ;
}

//-----------------------------------------------------------------------------
// OpenGL
//-----------------------------------------------------------------------------
/*!
 * xyz`(x:,y:,z:)
 * @param[in] len ̒
 */
inline int DrawAxis(double len, double line_width = 5.0)
{
	glLineWidth(line_width);

	// x
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glEnd();

	// y
	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, len, 0.0);
	glEnd();

	// z
	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, 0.0, len);
	glEnd();

	return 1;
}

/*!
 * ~`
 * @param[in] rad,len aƒS
 * @param[in] slices  |Sߎۂ̕
 */
static void DrawCylinder(double rad, double len, int up, int slices)
{
	GLUquadricObj *qobj;
	qobj = gluNewQuadric();

	glPushMatrix();
	switch(up){
	case 0:
		glRotatef(-90.0, 0.0, 1.0, 0.0);
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	case 1:
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	case 2:
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	default:
		glTranslatef(0.0, 0.0, -0.5*len);
	}

	gluQuadricDrawStyle(qobj, GLU_FILL);
	gluQuadricNormals(qobj, GLU_SMOOTH);
	gluCylinder(qobj, rad, rad, len, slices, slices);

	glPushMatrix();
	glRotatef(180.0, 1.0, 0.0, 0.0);
	gluDisk(qobj, 0.0, rad, slices, slices);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0.0, 0.0, len);
	gluDisk(qobj, 0.0, rad, slices, slices);
	glPopMatrix();

	glPopMatrix();
}

/*!
 * JvZ`(~̗[ɔ`)
 * @param[in] rad,len aƒS
 * @param[in] slices  |Sߎۂ̕
 */
static void DrawCapsule(double rad, double len, int up, int slices)
{
	GLUquadricObj *qobj;
	qobj = gluNewQuadric();

	glPushMatrix();
	switch(up){
	case 0:
		glRotatef(-90.0, 0.0, 1.0, 0.0);
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	case 1:
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	case 2:
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	default:
		glTranslatef(0.0, 0.0, -0.5*len);
	}

	gluQuadricDrawStyle(qobj, GLU_FILL);
	gluQuadricNormals(qobj, GLU_SMOOTH);
	gluCylinder(qobj, rad, rad, len, slices, slices);

	glPushMatrix();
	glutSolidSphere(rad, slices, slices);
	glPopMatrix();

	glPushMatrix();
	glTranslatef(0.0, 0.0, len);
	glutSolidSphere(rad, slices, slices);
	glPopMatrix();

	glPopMatrix();

}

/*!
 * ~`(~̗[ɔ`)
 * @param[in] rad,len aƒS
 * @param[in] slices  |Sߎۂ̕
 */
static void DrawCone(double rad, double len, int up, int slices)
{
	glPushMatrix();
	switch(up){
	case 0:
		glRotatef(-90.0, 0.0, 1.0, 0.0);
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	case 1:
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	case 2:
		glTranslatef(0.0, 0.0, -0.5*len);
		break;
	default:
		glTranslatef(0.0, 0.0, -0.5*len);
	}

	glutSolidCone(rad, len, slices, slices);

	glPopMatrix();

}



#endif // #ifndef _RX_UTILS_H_